#include "burp.h"
#include "prog.h"
#include "msg.h"
#include "rs_buf.h"
#include "lock.h"
#include "handy.h"
#include "asyncio.h"
#include "counter.h"

static int manifestend=0;

#if defined(WIN32_VSS)
#include "vss.h"

static int start_vss(struct config *conf)
{
	int errors=0;
	//char buf[256]="";

	if(g_pVSSClient->InitializeForBackup())
	{
		int i=0;
		int j=0;
		/* tell vss which drives to snapshot */
		char szWinDriveLetters[27];
		for(i=0, j=0; i < conf->sdcount && j<26; i++)
		{
			const char *path=conf->startdir[i]->path;
			if(strlen(path)>2 && isalpha(path[0]) && path[1]==':')
			{
				int x=0;
				// Try not to add the same letter twice.
				for(x=0; x<j; x++)
				  if(toupper(path[0])==szWinDriveLetters[x])
					break;
				if(x<j) continue;
				szWinDriveLetters[j++]=toupper(path[0]);
			}
		}
		szWinDriveLetters[j]='\0';
		printf("Generate VSS snapshots.\n");
		printf("Driver=\"%s\", Drive(s)=\"%s\"\n",
			g_pVSSClient->GetDriverName(),
			szWinDriveLetters);
		if(!g_pVSSClient->CreateSnapshots(szWinDriveLetters))
		{
			logp("Generate VSS snapshots failed.\n");
			errors++;
		}
		else
		{
			int i;
			for(i=0; i<(int)strlen(szWinDriveLetters); i++)
			{
			  logp("VSS drive letters: %d\n", i);
			  if(islower(szWinDriveLetters[i]))
			  {
				logp(_("Generate VSS snapshot of drive \"%c:\\\" failed. VSS support is disabled on this drive.\n"), szWinDriveLetters[i]);
				errors++;
			  }
			}

			for(i=0; i<(int)g_pVSSClient->GetWriterCount(); i++)
			{
			  logp("VSS writer count: %d\n", i);
			  if(g_pVSSClient->GetWriterState(i)<1)
			  {
				logp("VSS Writer (PrepareForBackup): %s\n", g_pVSSClient->GetWriterInfo(i));
				errors++;
			  }
			}
		}
	}
	else
	{
		berrno be;
		logp("VSS was not initialized properly.\n");
		logp("VSS support is disabled. ERR=%s\n",
			be.bstrerror(b_errno_win32));
		errors++;
	}

	return errors;
}

static int stop_vss(void)
{
	int errors=0;

	if(g_pVSSClient->CloseBackup())
	{
		int i=0;
		for(i=0; i<(int)g_pVSSClient->GetWriterCount(); i++)
		{
			if(g_pVSSClient->GetWriterState(i)<1)
				errors++;
			logp("VSS Writer (BackupComplete): %s\n",
				g_pVSSClient->GetWriterInfo(i));
		}
	}
	Win32ConvCleanupCache();
	return errors;
}
char *unix_name_to_win32(char *name);

#endif // WIN32_VSS

static int send_new_file(int type, const char *path)
{
	// send the whole file
	// only open the file if it has stuff in it
#if defined(WIN32_VSS)
	if((type==FT_DIRBEGIN && send_whole_file_gz(path, NULL))) return -1;
#endif
	if((type==FT_REG || type==FT_REGE)
	  && send_whole_file_gz(path, NULL)) return -1;
	//logp("send_new_file finish %s\n", path);
	if(async_write_str('x', "endfile")) // finish file
		return -1;
	return 0;
}

static int send_dpth(const char *fname, const char *dpth)
{
	if(async_write_str('t', dpth))
	{
		logp("could not send dpth for: %s (%s)\n", fname, dpth);
		return -1;
	}
	return 0;
}

static int send_attribs(const char *fname, const char *attribs)
{
	if(async_write_str('r', attribs))
	{
		logp("could not send attribs for: %s\n", fname);
		return -1;
	}
	return 0;
}

static int send_fname(const char *fname)
{
	if(async_write_str('f', fname))
	{
		logp("could not send filename for: %s\n", fname);
		return -1;
	}
	return 0;
}

int send_Fname(const char *fname)
{
        if(async_write_str('F', fname))
        {
                logp("could not send filename for: %s\n", fname);
                return -1;
        }
        return 0;
}

static int send_attribs_and_fname(const char *attribs, const char *fname)
{
	if(send_attribs(fname, attribs)
	  || send_fname(fname))
		return -1;
	return 0;
}

static int send_attribs_and_Fname(const char *attribs, const char *fname)
{
	if(send_attribs(fname, attribs)
	  || send_Fname(fname))
		return -1;
	return 0;
}

static void free_cbufs(char **rpath, char **sstat, char **spath)
{
	if(*rpath) { free(*rpath); *rpath=NULL; }
	if(*sstat) { free(*sstat); *sstat=NULL; }
	if(*spath) { free(*spath); *spath=NULL; }
}

#define DO_UNCHANGED	3
#define DO_DIFF_STUFF	2
#define DO_WHOLE_FILE	1

static int check_cbufs(FF_PKT *ff, char **cspath, char **csstat, char **ctpath, struct stat *statp)
{
	int pcmp;

	if((pcmp=pathcmp(ff->fname, *cspath))==0)
	{
		//logp("match\n");
		// found it - need to do diff/patching
		if(statp->st_mtime!=ff->statp.st_mtime
		  || statp->st_ctime!=ff->statp.st_ctime)
			return DO_DIFF_STUFF;
		else
			return DO_UNCHANGED;
	}
	else if(pcmp<0)
	{
		// server ahead - send a new file
		//logp("ahead\n");
		return DO_WHOLE_FILE;
	}
	// server behind - need to keep reading
	//logp("behind: '%s' '%s'\n", ff->fname, *cspath);
	free_cbufs(cspath, csstat, ctpath);
	return 0;
}

static int load_signature_and_send_delta(const char *rpath)
{
	rs_job_t *job;
	rs_result r;
	rs_signature_t *sumset=NULL;
#ifdef HAVE_WIN32
        BFILE bfd;
#else
	FILE *in=NULL;
#endif
	rs_filebuf_t *infb=NULL;
	rs_filebuf_t *outfb=NULL;
	rs_buffers_t rsbuf;
	memset(&rsbuf, 0, sizeof(rsbuf));

//logp("loadsig %s\n", rpath);
/* handy debug - write the sig to a file.
	FILE *xp=NULL;
	xp=fopen("/tmp/test.gz", "w");
	transfer_file_in(NULL, xp, NULL);
	fclose(xp);
	exit(1);
*/


	job = rs_loadsig_begin(&sumset);
	if((r=do_rs_run(job, NULL, NULL, NULL, NULL, NULL, async_get_fd(), -1, ZTYPE_INFLATE, ZTYPE_NONE)))
	{
		rs_free_sumset(sumset);
		return r;
	}
	if((r=rs_build_hash_table(sumset)))
	{
		rs_free_sumset(sumset);
		return r;
	}
	rs_job_free(job);
	job=NULL;
//logp("end loadsig\n");
//logp("\n");

#ifdef HAVE_WIN32
        binit(&bfd);
        if(bopen(&bfd, rpath, O_RDONLY | O_BINARY | O_NOATIME, S_IRUSR | S_IWUSR)<0)
        {
                berrno be;
                logp("Could not open %s: %s\n", rpath, be.bstrerror(errno));
                return -1;
        }
#else
	//logp("opening: %s\n", rpath);
	if(!(in=fopen(rpath, "rb")))
	{
		logp("could not open '%s' in order to generate delta.\n",
			rpath);
		rs_free_sumset(sumset);
		return RS_IO_ERROR;
	}
#endif
//logp("start delta\n");

	if(!(job=rs_delta_begin(sumset)))
	{
		logp("could not start delta job.\n");
		rs_free_sumset(sumset);
		return RS_IO_ERROR;
	}

#ifdef HAVE_WIN32
	if(!(infb=rs_filebuf_new(&bfd, NULL, NULL, -1, ZTYPE_NONE, rs_inbuflen))
	  || !(outfb=rs_filebuf_new(NULL, NULL, NULL, async_get_fd(), ZTYPE_DEFLATE, rs_outbuflen)))
	{
		logp("could not rs_filebuf_new for delta\n");
		if(infb) rs_filebuf_free(infb);
		return -1;
	}
#else
	if(!(infb=rs_filebuf_new(NULL, in, NULL, -1, ZTYPE_NONE, rs_inbuflen))
	  || !(outfb=rs_filebuf_new(NULL, NULL, NULL, async_get_fd(), ZTYPE_DEFLATE, rs_outbuflen)))
	{
		logp("could not rs_filebuf_new for delta\n");
		if(infb) rs_filebuf_free(infb);
		return -1;
	}
#endif
//logp("start delta loop\n");

	while(1)
	{
		size_t wlen=0;
		rs_result delresult;
		delresult=rs_async(job, &rsbuf, infb, outfb);
		if(delresult==RS_DONE)
		{
			r=delresult;
			//logp("delresult done\n");
			break;
		}
		else if(delresult==RS_BLOCKED || delresult==RS_RUNNING)
		{
			//logp("delresult running/blocked: %d\n", delresult);
			// Keep going
		}
		else
		{
			logp("error in rs_async for delta: %d\n", delresult);
			r=delresult;
			break;
		}
		// FIX ME: get it to read stuff (errors, for example) here too.
		if(async_rw(NULL, NULL, '\0', '\0', NULL, &wlen))
			return -1;
	}

	if(r!=RS_DONE)
		logp("delta loop returned: %d\n", r);

//logp("after delta loop: %d\n", r);
//logp("\n");

	rs_filebuf_free(infb);
	rs_filebuf_free(outfb);
	rs_job_free(job);
	rs_free_sumset(sumset);
#ifdef HAVE_WIN32
        bclose(&bfd);
#else
	if(in) fclose(in);
#endif

	if(r==RS_DONE && async_write_str('x', "endfile")) // finish delta file
		return -1;

	//logp("end of load_sig_send_delta for: %s\n", rpath);

	return r;
}

static int do_write_sequence(FF_PKT *ff, const char *attribs, int writing, const char *rpath, const char *sstat, const char *spath)
{
	if(writing==DO_UNCHANGED)
	{
		//logp("unchanged file: %s\n", ff->fname);
		do_filecounter('f', 1);
		if(send_dpth(ff->fname, spath))
			return -1;
		if(send_attribs_and_fname(attribs, ff->fname))
			return -1;
	}
	else if(writing==DO_WHOLE_FILE)
	{
		//logp("new file: %s\n", ff->fname);
		do_filecounter('F', 1);
		if(send_attribs_and_Fname(attribs, ff->fname))
			return -1;
		if(send_new_file(ff->type, ff->fname))
			return -1;
	}
	else if(writing==DO_DIFF_STUFF)
	{
		//logp("sig/delta file: %s\n", ff->fname);
		do_filecounter('x', 1);

		if(send_dpth(ff->fname, spath))
			return -1;
		if(send_attribs_and_Fname(attribs, ff->fname))
			return -1;
		if(load_signature_and_send_delta(ff->fname)) return -1;
	}

	return 0;
}

static int backup_read_write(FF_PKT *ff, const char *attribs)
{
	int writing=0;
	static char *rpath=NULL; // current real path
	static char *sstat=NULL; // current stat
	static char *spath=NULL; // current stored path
	static struct stat statp; // current stat in a struct

	//logp("do file: %s\n", ff->fname);

	if(rpath && sstat && spath)
		writing=check_cbufs(ff, &rpath, &sstat, &spath, &statp);

	while(1)
	{
		if(manifestend)
		{
			//logp("manifest ended - write whole file\n");
			writing=DO_WHOLE_FILE;
		}
		if(writing)
		{
			if(do_write_sequence(ff, attribs,
				writing, rpath, sstat, spath))
			{
				free_cbufs(&rpath, &sstat, &spath);
				logp("error in do_write_sequence\n");
				return -1;
			}
			if(writing==DO_DIFF_STUFF || writing==DO_UNCHANGED)
				free_cbufs(&rpath, &sstat, &spath);
			return 0;
		}
		else
		{
			char rcmd='\0';
			char *rbuf=NULL; // read buffer
			size_t rlen=0; // read buffer length
			size_t wlen=0;
			if(async_rw(&rcmd, &rbuf, &rlen, '\0', NULL, &wlen))
			{
				logp("error in async_rw\n");
				return -1;
			}
			if(!rcmd) continue;
			//logp("NOW: rcmd '%c'\n", rcmd);
			if(rcmd=='a' || rcmd=='x')
			{
				// skip
				//logp("rcmd '%c'\n", rcmd);
				free(rbuf);
				continue;
			}
			else if(rcmd=='p')
			{
				free_cbufs(&rpath, &sstat, &spath);
				rpath=rbuf;
				//logp("got rpath: %s\n", rpath);
			}
			else if(rcmd=='r')
			{
				if(!rpath)
				{
					logp("unexpected manifest stat cmd from server: %c %s\n", rcmd, rbuf);
					free(rbuf);
					free_cbufs(&rpath, &sstat, &spath);
					return -1;
				}
				sstat=rbuf;
				decode_stat(sstat, &statp);
				//logp("got sstat: %s\n", sstat);
			}
			else if(rcmd=='t')
			{
				if(!rpath || !sstat)
				{
					logp("unexpected manifest tpath cmd from server: %c %s\n", rcmd, rbuf);
					free(rbuf);
					free_cbufs(&rpath, &sstat, &spath);
					return -1;
				}
				spath=rbuf;
			}
			else if(rcmd=='c' && !strcmp(rbuf, "okmanifestend"))
			{
				free_cbufs(&rpath, &sstat, &spath);
				free(rbuf);
				manifestend++;
				//logp("manifest ended - write whole file\n");
				writing=DO_WHOLE_FILE;
			}
			else
			{
				logp("unexpected manifest cmd from server: %c %s\n", rcmd, rbuf);
				free_cbufs(&rpath, &sstat, &spath);
				free(rbuf);
				return -1;
			}

			if(rpath && sstat && spath)
			{
				writing=check_cbufs(ff,
					&rpath, &sstat, &spath, &statp);
			}
		}
	}
	return 0;
}

static int backup_file(FF_PKT *ff, struct config *conf)
{
	int ret=0;
	char attribs[MAXSTRING];

	encode_stat(attribs, &ff->statp);

	ret=backup_read_write(ff, attribs);
	if(ret)
	{
		logp("error in backup_read_write\n");
		return -1;
	}
	return 0;
}

static int send_file(FF_PKT *ff, bool top_level, struct config *conf)
{
   char msg[128]="";
   char attribs[MAXSTRING];

   if(!file_is_included(conf->incexcdir, conf->iecount, ff->fname))
   {
	//logp("skip: %s\n", ff->fname);
	return 0;
   }

   //logp("%d: %s\n", ff->type, ff->fname);

   switch (ff->type) {
   case FT_LNKSAVED:
        //printf("Lnka: %s -> %s\n", ff->fname, ff->link);
   	encode_stat(attribs, &ff->statp);
	if(async_write_str('r', attribs)
	  || async_write_str('L', ff->fname)
	  || async_write_str('L', ff->link))
		return -1;
	do_filecounter('L', 1);
      break;
   case FT_REGE:
   case FT_REG:
      return backup_file(ff, conf);
      break;
   case FT_LNK:
	//printf("link: %s -> %s\n", ff->fname, ff->link);
   	encode_stat(attribs, &ff->statp);
	if(async_write_str('r', attribs)
	  || async_write_str('l', ff->fname)
	  || async_write_str('l', ff->link))
		return -1;
	do_filecounter('l', 1);
      break;
   case FT_DIREND:
      return 0;
   case FT_NORECURSE:
   case FT_NOFSCHG:
   case FT_INVALIDFS:
   case FT_INVALIDDT:
   case FT_DIRBEGIN:
	{
         char errmsg[100] = "";
         if (ff->type == FT_NORECURSE)
	    snprintf(errmsg, sizeof(errmsg), _("\t[will not descend: recursion turned off]"));
         else if (ff->type == FT_NOFSCHG)
            snprintf(errmsg, sizeof(errmsg), _("\t[will not descend: file system change not allowed]"));
         else if (ff->type == FT_INVALIDFS)
            snprintf(errmsg, sizeof(errmsg), _("\t[will not descend: disallowed file system]"));
         else if (ff->type == FT_INVALIDDT)
            snprintf(errmsg, sizeof(errmsg), _("\t[will not descend: disallowed drive type]"));
	 if(*errmsg)
	 {
		snprintf(msg, sizeof(msg),
			"%s%s%s\n", "Dir: ", ff->fname, errmsg);
		logw(msg);
	 }
	 else
	 {
#if defined(WIN32_VSS)
     		return backup_file(ff, conf);
#else
		encode_stat(attribs, &ff->statp);
		if(async_write_str('r', attribs)
		  || async_write_str('d', ff->fname))
			return -1;
		do_filecounter('d', 1);
#endif
	 }
	}
      break;
   case FT_SPEC: // special file - fifo, socket, device node...
      encode_stat(attribs, &ff->statp);
      if(async_write_str('r', attribs)
	  || async_write_str('s', ff->fname))
		return -1;
      do_filecounter('s', 1);
      break;
   case FT_NOACCESS:
      logw(_("Err: Could not access %s: %s"), ff->fname, strerror(errno));
      break;
   case FT_NOFOLLOW:
      logw(_("Err: Could not follow ff->link %s: %s"), ff->fname, strerror(errno));
      break;
   case FT_NOSTAT:
      logw(_("Err: Could not stat %s: %s"), ff->fname, strerror(errno));
      break;
   case FT_NOCHG:
      logw(_("Skip: File not saved. No change. %s"), ff->fname);
      break;
   case FT_ISARCH:
      logw(_("Err: Attempt to backup archive. Not saved. %s"), ff->fname);
      break;
   case FT_NOOPEN:
      logw(_("Err: Could not open directory %s: %s"), ff->fname, strerror(errno));
      break;
   case FT_REPARSE:
      logw(_("Err: Windows reparse point: %s"), ff->fname);
      break;
   case FT_RAW:
      logw(_("Err: Raw partition: %s"), ff->fname);
      break;
   case FT_FIFO:
      logw(_("Err: Fifo: %s"), ff->fname);
      break;
   default:
      logw(_("Err: Unknown file ff->type %d: %s"), ff->type, ff->fname);
      break;
   }
/*
   if (attrs) {
      char attr[200];
      encode_attribsEx(NULL, attr, ff);
      if (*attr != 0) {
         printf("AttrEx=%s\n", attr);
      }
   }
*/
   return 0;
}

#if defined(HAVE_WIN32)

static int enable_priv(HANDLE hToken, const char *name, int ignore_errors)
{
    TOKEN_PRIVILEGES tkp;
    DWORD lerror;

    if (!(p_LookupPrivilegeValue && p_AdjustTokenPrivileges)) {
       return 0;                      /* not avail on this OS */
    }

    // Get the LUID for the security privilege.
    if (!p_LookupPrivilegeValue(NULL, name,  &tkp.Privileges[0].Luid)) {
	logp("LookupPrivilegeValue: %s\n", GetLastError());
       return 0;
    }

    /* Set the security privilege for this process. */
    tkp.PrivilegeCount = 1;
    tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
    p_AdjustTokenPrivileges(hToken, FALSE, &tkp, 0, NULL, NULL);
    lerror = GetLastError();
    if (lerror != ERROR_SUCCESS) {
       if (!ignore_errors) {
	  logp("AdjustTokenPrivileges set %s: %s\n", name, lerror);
       }
       return 0;
    }
    return 1;
}

/*
 * Setup privileges we think we will need.  We probably do not need
 *  the SE_SECURITY_NAME, but since nothing seems to be working,
 *  we get it hoping to fix the problems.
 */
static int enable_backup_privileges(int ignore_errors)
{
    HANDLE hToken, hProcess;
    int stat = 0;

    if (!p_OpenProcessToken) {
       return 0;                      /* No avail on this OS */
    }

    hProcess = OpenProcess(PROCESS_ALL_ACCESS, FALSE, GetCurrentProcessId());

    // Get a token for this process.
    if (!p_OpenProcessToken(hProcess,
            TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken)) {
       if (!ignore_errors) {
          logp("OpenProcessToken: %s\n", GetLastError());
       }
       /* Forge on anyway */
    }

    /* Return a bit map of permissions set. */
    if (enable_priv(hToken, SE_BACKUP_NAME, ignore_errors)) {
       stat |= 1<<1;
    }
    if (enable_priv(hToken, SE_RESTORE_NAME, ignore_errors)) {
       stat |= 1<<2;
    }
#if 0
    if (enable_priv(hToken, SE_SECURITY_NAME, ignore_errors)) {
       stat |= 1<<0;
    }
    if (enable_priv(hToken, SE_TAKE_OWNERSHIP_NAME, ignore_errors)) {
       stat |= 1<<3;
    }
    if (enable_priv(hToken, SE_ASSIGNPRIMARYTOKEN_NAME, ignore_errors)) {
       stat |= 1<<4;
    }
    if (enable_priv(hToken, SE_SYSTEM_ENVIRONMENT_NAME, ignore_errors)) {
       stat |= 1<<5;
    }
    if (enable_priv(hToken, SE_CREATE_TOKEN_NAME, ignore_errors)) {
       stat |= 1<<6;
    }
    if (enable_priv(hToken, SE_MACHINE_ACCOUNT_NAME, ignore_errors)) {
       stat |= 1<<7;
    }
    if (enable_priv(hToken, SE_TCB_NAME, ignore_errors)) {
       stat |= 1<<8;
    }
    if (enable_priv(hToken, SE_CREATE_PERMANENT_NAME, ignore_errors)) {
       stat |= 1<<10;
    }
#endif
    if (stat) {
       stat |= 1<<9;
    }

    CloseHandle(hToken);
    CloseHandle(hProcess);
    return stat;
}
#endif  /* HAVE_WIN32 */

void close_fd(int *fd)
{
	if(*fd>0)
	{
		//logp("closing %d\n", *fd);
		close(*fd);
		*fd=-1;
	}
}

static int init_socket(const char *host, int port)
{
        int rfd=-1;
        struct sockaddr_in sin;
        struct hostent *hp=NULL;

        if(!(hp=gethostbyname(host)))
        {
                logp("unknown host: %s\n", host);
                return -1;
        }

        memset(&sin, 0, sizeof(sin));
        sin.sin_family=AF_INET;
        memcpy((char *)&sin.sin_addr, hp->h_addr, hp->h_length);
        sin.sin_port=htons(port);

        if((rfd=socket(PF_INET, SOCK_STREAM, 0))<0)
        {
		berrno be;
                logp("socket error: %s\n", be.bstrerror());
                return -1;
        }
        //setsockopt(rfd, IPPROTO_TCP, TCP_NODELAY, (char *) &flag, sizeof(int));
        if(connect(rfd, (struct sockaddr *)&sin, sizeof(sin))<0)
        {
		berrno be;
                logp("connect error, %s: %s\n", host, be.bstrerror(b_errno_win32));
                close_fd(&rfd);
                return -1;
        }
	set_non_blocking(rfd);

#ifdef HAVE_WIN32
	setmode(rfd, O_BINARY);
#endif
        return rfd;
}

static char *encode_mode(mode_t mode, char *buf)
{
  char *cp = buf;

  *cp++ = S_ISDIR(mode) ? 'd' : S_ISBLK(mode)  ? 'b' : S_ISCHR(mode)  ? 'c' :
          S_ISLNK(mode) ? 'l' : S_ISFIFO(mode) ? 'f' : S_ISSOCK(mode) ? 's' : '-';
  *cp++ = mode & S_IRUSR ? 'r' : '-';
  *cp++ = mode & S_IWUSR ? 'w' : '-';
  *cp++ = (mode & S_ISUID
               ? (mode & S_IXUSR ? 's' : 'S')
               : (mode & S_IXUSR ? 'x' : '-'));
  *cp++ = mode & S_IRGRP ? 'r' : '-';
  *cp++ = mode & S_IWGRP ? 'w' : '-';
  *cp++ = (mode & S_ISGID
               ? (mode & S_IXGRP ? 's' : 'S')
               : (mode & S_IXGRP ? 'x' : '-'));
  *cp++ = mode & S_IROTH ? 'r' : '-';
  *cp++ = mode & S_IWOTH ? 'w' : '-';
  *cp++ = (mode & S_ISVTX
               ? (mode & S_IXOTH ? 't' : 'T')
               : (mode & S_IXOTH ? 'x' : '-'));
  *cp = '\0';
  return cp;
}

static char *encode_time(utime_t utime, char *buf)
{
   const struct tm *tm;
   int n = 0;
   time_t time = utime;

#if defined(HAVE_WIN32)
   /*
    * Avoid a seg fault in Microsoft's CRT localtime_r(),
    *  which incorrectly references a NULL returned from gmtime() if
    *  time is negative before or after the timezone adjustment.
    */
   struct tm *gtm;

   if ((gtm = gmtime(&time)) == NULL) {
      return buf;
   }

   if (gtm->tm_year == 1970 && gtm->tm_mon == 1 && gtm->tm_mday < 3) {
      return buf;
   }
#endif

   if((tm=localtime(&time)))
   {
      n = sprintf(buf, "%04d-%02d-%02d %02d:%02d:%02d",
                   tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday,
                   tm->tm_hour, tm->tm_min, tm->tm_sec);
   }
   return buf+n;
}


static void ls_output(char *buf, const char *fname, struct stat *statp)
{
	char *p;
	const char *f;
	int n;
	time_t time;

	p = encode_mode(statp->st_mode, buf);
	n = sprintf(p, " %2d ", (uint32_t)statp->st_nlink);
	p += n;
	n = sprintf(p, "%4d %4d",
		(uint32_t)statp->st_uid,
		(uint32_t)statp->st_gid);
	p += n;
	n = sprintf(p, " %7lu ", (unsigned long) statp->st_size);
	p += n;
	if (statp->st_ctime > statp->st_mtime)
		time = statp->st_ctime;
	else
		time = statp->st_mtime;

	/* Display most recent time */
	p = encode_time(time, p);
	*p++ = ' ';
	for (f=fname; *f; )
		*p++ = *f++;
	*p = 0;
}

static int do_list(const char *backup, const char *listregex, enum action act)
{
	int ret=0;
	size_t slen=0;
	char msg[64]="";
	char scmd;
	struct stat statp;
	char *statbuf=NULL;
	char ls[256]="";
	char *dpth=NULL;
//logp("in do_list\n");

	snprintf(msg, sizeof(msg), "list %s:%s", backup?:"", listregex?:"");
	if(async_write_str('c', msg)
	  || async_read_expect('c', "ok"))
		return -1;

	while(1)
	{
		char fcmd;
		size_t flen=0;
		char *fname=NULL;
		if(async_read(&scmd, &statbuf, &slen))
		{
			ret=-1; break;
		}
		if(scmd=='b')
		{
			// A backup timestamp, just print it.
			logp("Backup: %s\n", statbuf);
			if(listregex) printf("With regex: %s\n", listregex);
			if(statbuf) { free(statbuf); statbuf=NULL; }
			continue;
		}
		else if(scmd=='t')
		{
			if(statbuf) { free(statbuf); statbuf=NULL; }
			continue;
		}
		else if(scmd!='r')
		{
			logp("expected 'r' cmd - got %c:%s\n", scmd, statbuf);
			ret=-1; break;
		}

		decode_stat(statbuf, &statp);

		if(async_read(&fcmd, &fname, &flen))
		{
			logp("got stat without an object\n");
			ret=-1; break;
		}
		else if(fcmd=='d' || fcmd=='f' || fcmd=='s')
		{
			*ls='\0';
			if(act==ACTION_LONG_LIST)
			{
				ls_output(ls, fname, &statp);
				printf("%s\n", ls);
			}
			else
			{
				printf("%s\n", fname);
			}
		}
		else if(fcmd=='l' || fcmd=='L') // symlink or hardlink
		{
			char lcmd;
			size_t llen=0;
			char *lname=NULL;
			if(async_read(&lcmd, &lname, &llen)
			  || lcmd!=fcmd)
			{
				logp("could not get second part of link %c:%s\n", fcmd, fname);
				ret=-1;
			}
			else
			{
				if(act==ACTION_LONG_LIST)
				{
					*ls='\0';
					ls_output(ls, fname, &statp);
					printf("%s -> %s\n", ls, lname);
				}
				else
				{
					printf("%s\n", fname);
				}
			}
			if(lname) free(lname);
		}
		if(fname) free(fname);
		if(statbuf) { free(statbuf); statbuf=NULL; }
	}
	if(statbuf) free(statbuf);
	if(dpth) free(dpth);
	return ret;
}

static int do_backup(struct config *conf)
{
	int sd=0;
	int ret=0;
	FF_PKT *ff=NULL;

	if(async_write_str('c', "backup")
	  || async_read_expect('c', "ok")
	// Send our starting point, so the server can forward through the
	// manifest.
	  || async_write_str('c', conf->startdir[sd]->path)
	  || async_read_expect('c', "ok"))
		return -1;

#if defined(HAVE_WIN32)
	enable_backup_privileges(1 /* ignore_errors */);
#endif
#if defined(WIN32_VSS)
	start_vss(conf);
#endif
	ff=init_find_files();
	reset_filecounter();
	for(; sd < conf->sdcount; sd++)
	{
		if(conf->startdir[sd]->include)
		{
			if((ret=find_files(ff, conf,
				conf->startdir[sd]->path, send_file)))
					break;
		}
	}
	end_filecounter_backup(1);
	term_find_files(ff);
#if defined(WIN32_VSS)
	stop_vss();
#endif
	if(ret)
	{
		logp("error in find_files\n");
	}
	else
	{
		char cmd;
		int quit=0;
		size_t len=0;
		char *buf=NULL;

		if(async_write_str('c', "backupend")) return -1;

		while(!quit && !async_read(&cmd, &buf, &len))
		{
			if(cmd=='c')
			{
				if(!strcmp(buf, "okmanifestend")) { }
				else if(!strcmp(buf, "okbackupend"))
				{
					logp("backup finished\n");
					ret=0; quit++;
				}
			}
			else if(cmd=='e')
			{
				logp("wanted okmanifestend - got: %c:%s\n",
					cmd, buf);
				ret=-1; quit++;
			}
			free(buf);
			buf=NULL;
		}
	}
	return ret;
}

static int say_hello(struct config *conf)
{
	if(async_write_str('c', "hello")
	  || async_read_expect('c', "whoareyou")
	  || async_write_str('c', conf->cname)
	  || async_read_expect('c', "okpassword")
	  || async_write_str('c', conf->password)
	  || async_read_expect('c', "ok"))
		return -1;
	//logp("auth ok\n");
	return 0;
}

static int restore_interrupt(char fcmd, const char *dpth, const char *msg)
{
	int ret=0;
	int quit=0;
	char *buf=NULL;

	do_filecounter('w', 0);
	logp("WARNING: %s\n", msg);
	if(async_write_str('w', msg)) return -1;

	if(fcmd=='l' || fcmd=='L')
	{
		char cmd;
		size_t len=0;
		char *buf=NULL;
		// It is a link - read the next line before returning.
		if(async_read(&cmd, &buf, &len))
		{
			if(buf) free(buf);
			return -1;
		}
		if(cmd!='l' && cmd!='L')
		{
			logp("expected link command when flushing - got: %c\n",
				cmd);
			if(buf) free(buf);
			return -1;
		}
		if(buf) free(buf);
	}

	// If it is file data, get the server
	// to interrupt the flow and move on.
	if(fcmd!='f' || !dpth) return 0;
	if(async_write_str('i', dpth))
	{
		ret=-1;
		quit++;
	}
	// Read to the end file marker.
	while(!quit)
	{
		size_t len=0;
		char cmd='\0';
		if(async_read(&cmd, &buf, &len))
		{
			ret=-1; quit++;
		}
		if(!ret && len)
		{
		  if(cmd=='a')
		  {
			continue;
		  }
		  else if(cmd=='x')
		  {
			break;
		  }
		  else
		  {
			logp("unexpected cmd from server while flushing: %c:%s\n", cmd, buf);
			ret=-1; quit++;
		  }
		}
		if(buf) { free(buf); buf=NULL; }
	}
	if(buf) free(buf);
	return ret;
}

static int make_link(const char *fname, const char *lnk, char cmd, const char *restoreprefix)
{
	int ret=-1;

#ifdef HAVE_WIN32
	logp("windows seems not to support hardlinks or symlinks\n");
#else
	unlink(fname);
	if(cmd=='L')
	{
		char *flnk=NULL;
		if(!(flnk=prepend_s(restoreprefix, lnk, strlen(lnk))))
		{
			logp("out of memory\n");
			return -1;
		}
		printf("%s -> %s\n", fname, flnk);
		ret=link(flnk, fname);
		free(flnk);
	}
	else
	{
		printf("%s -> %s\n", fname, lnk);
		ret=symlink(lnk, fname);
	}
#endif

	if(ret) logp("could not %slink %s -> %s\n", cmd=='L'?"hard":"sym",
		fname, cmd=='L'?lnk:fname);

	return ret;
}

static int restore_file(char cmd, const char *fname, struct stat *statp)
{
	size_t len=0;
	int ret=0;
	char *rpath=NULL;
#ifdef HAVE_WIN32
	int bopenret;
	BFILE bfd;
#else
	FILE *fp=NULL;
#endif

	if(build_path(fname, "", len, &rpath))
	{
		char msg[256]="";
		// failed - do a warning
		snprintf(msg, sizeof(msg), "build path failed: %s", rpath);
		if(restore_interrupt(cmd, NULL, msg))
			ret=-1;
	}

	logp("restore file: %s\n", rpath);
#ifdef HAVE_WIN32
	binit(&bfd);
	set_win32_backup(&bfd);
	if(S_ISDIR(statp->st_mode))
	{
		mkdir(rpath, 0777);
		bopenret=bopen(&bfd, rpath, O_WRONLY | O_BINARY, 0);
	}
	else
		bopenret=bopen(&bfd, rpath,
		  O_WRONLY | O_CREAT | O_TRUNC | O_BINARY, S_IRUSR | S_IWUSR);
	if(bopenret<0)
	{
		berrno be;
		char msg[256]="";
		snprintf(msg, sizeof(msg),
			"Could not open for writing %s: %s",
				rpath, be.bstrerror(errno));
		if(restore_interrupt(cmd, NULL, msg))
			ret=-1;
	}
#else
	if(!(fp=open_file(rpath, "wb")))
	{
		char msg[256]="";
		snprintf(msg, sizeof(msg),
			"Could not open for writing %s: %s",
				rpath, strerror(errno));
		if(restore_interrupt(cmd, NULL, msg))
			ret=-1;
	}
#endif

	if(!ret)
	{
#ifdef HAVE_WIN32
		ret=transfer_gzfile_in(&bfd, NULL, NULL);
		bclose(&bfd);
#else
		ret=transfer_gzfile_in(NULL, fp, NULL);
		close_fp(&fp);
		if(!ret) set_attributes(rpath, cmd, statp);
#endif
		if(ret)
		{
			char msg[256]="";
			snprintf(msg, sizeof(msg),
				"Could not transfer file in: %s",
					rpath);
			if(restore_interrupt(cmd, NULL, msg))
				ret=-1;
		}
	}
	if(!ret) do_filecounter(cmd, 0);

	if(rpath) free(rpath);
	return ret;
}

static int restore_special(char cmd, const char *fname, struct stat *statp)
{
	int ret=0;
#ifdef HAVE_WIN32
	char msg[256]="";
	logw("Cannot restore special files to Windows: %s\n", fname);
#else
	char *rpath=NULL;
	if(build_path(fname, "", 0, &rpath))
	{
		char msg[256]="";
		// failed - do a warning
		snprintf(msg, sizeof(msg), "build path failed: %s", rpath);
		if(restore_interrupt(cmd, NULL, msg))
			ret=-1;
	}
	if(S_ISFIFO(statp->st_mode))
	{
		if(mkfifo(rpath, statp->st_mode) && errno!=EEXIST)
		{
			char msg[256]="";
			snprintf(msg, sizeof(msg),
				"Cannot make fifo: %s\n", strerror(errno));
			logw(msg);
		}
		else
		{
			set_attributes(rpath, 's', statp);
			do_filecounter('s', 0);
		}
	}
	else if(S_ISSOCK(statp->st_mode)) {
		char msg[256]="";
		snprintf(msg, sizeof(msg),
			"Skipping restore of socket: %s\n", fname);
		logw(msg);
#ifdef S_IFDOOR     // Solaris high speed RPC mechanism
	} else if (S_ISDOOR(statp->st_mode)) {
		char msg[256]="";
		snprintf(msg, sizeof(msg),
			"Skipping restore of door file: %s\n", fname);
		logw(msg);
#endif
#ifdef S_IFPORT     // Solaris event port for handling AIO
	} else if (S_ISPORT(statp->st_mode)) {
		char msg[256]="";
		snprintf(msg, sizeof(msg),
			"Skipping restore of event port file: %s\n", fname);
		logw(msg);
#endif
	} else {
            if(mknod(fname, statp->st_mode, statp->st_rdev) && errno!=EEXIST)
	    {
		char msg[256]="";
		snprintf(msg, sizeof(msg),
			"Cannot make node: %s\n", strerror(errno));
		logw(msg);
            }
	    else
	    {
		set_attributes(rpath, 's', statp);
		do_filecounter('s', 0);
	    }
         }
#endif
	return ret;
}

static int restore_dir(char cmd, const char *dname, struct stat *statp)
{
	int ret=0;
//#ifndef HAVE_WIN32
	char *rpath=NULL;

	if(build_path(dname, "", 0, &rpath))
	{
		char msg[256]="";
		// failed - do a warning
		snprintf(msg, sizeof(msg), "build path failed: %s", rpath);
		if(restore_interrupt(cmd, NULL, msg))
			ret=-1;
	}
	else if(!is_dir(rpath) && mkdir(rpath, 0777))
	{
		char msg[256]="";
		snprintf(msg, sizeof(msg), "mkdir error: %s",
			strerror(errno));
		// failed - do a warning
		if(restore_interrupt(cmd, NULL, msg))
			ret=-1;
	}
	else
	{
		set_attributes(rpath, cmd, statp);
	}
//#endif
	if(!ret) do_filecounter(cmd, 0);
	return ret;
}

static int restore_link(char fcmd, const char *fname, size_t flen, struct stat *statp, const char *restoreprefix)
{
	int ret=0;
	char lcmd;
	size_t llen=0;
	char *lname=NULL;

	// Read where the link points to.	
	if(async_read(&lcmd, &lname, &llen))
	{
		ret=-1;
	}
	else if(fcmd!=lcmd)
	{
		ret=-1;
		async_write_str('e', "link cmd mismatch");
	}
	else if(make_link(fname, lname, lcmd, restoreprefix))
	{
		// failed - do a warning
		if(restore_interrupt(fcmd, NULL, "could not create link"))
			ret=-1;
	}
	else
	{
		set_attributes(fname, fcmd, statp);
	}
	if(!ret) do_filecounter(fcmd, 0);
	if(lname) free(lname);
	return ret;
}

static void strip_invalid_characters(char **path)
{
#ifdef HAVE_WIN32
      char *ch = *path;
      if (ch[0] != 0 && ch[1] != 0) {
         ch += 2;
         while (*ch) {
            switch (*ch) {
            case ':':
            case '<':
            case '>':
            case '*':
            case '?':
            case '|':
               *ch = '_';
                break;
            }
            ch++;
         }
      }
#endif
}

static int do_restore(struct config *conf, const char *backup, const char *restoreprefix, const char *restoreregex, int forceoverwrite)
{
	char cmd;
	int ars=0;
	int ret=0;
	int quit=0;
	size_t len=0;
	char *buf=NULL;
	char msg[64]="";
	struct stat statp;
	char *statbuf=NULL;
	size_t slen=0;
	struct stat checkstat;

	logp("doing restore\n");
	reset_filecounter();

	snprintf(msg, sizeof(msg), "restore %s:%s",
		backup?:"", restoreregex?:"");
	if(async_write_str('c', msg)
	  || async_read_expect('c', "ok"))
		return -1;
	logp("doing restore confirmed\n");

#if defined(HAVE_WIN32)
	enable_backup_privileges(1 /* ignore_errors */);
#endif

	while(!quit)
	{
		char *dpth=NULL;
		char *fullpath=NULL;

		if((ars=async_read_stat(NULL, NULL, &statbuf, &slen, &statp, &dpth)))
		{
			if(ars<0) ret=-1; // error
			else
			{
				ret=0; // restore finished ok
				logp("got restore end\n");
				if(async_write_str('c', "restoreend ok"))
					ret=-1;
			}
			break;
		}
		if(async_read(&cmd, &buf, &len))
		{
			logp("async_read error\n");
			ret=-1;
			if(dpth) free(dpth);
			break;
		}
		switch(cmd)
		{
			case 'd':
			case 'f':
			case 'l':
			case 'L':
			case 's':
				if(!(fullpath=prepend_s(restoreprefix,
					buf, strlen(buf))))
				{
					log_and_send("out of memory");
					ret=-1;
					quit++;
				}
				strip_invalid_characters(&fullpath);
				if(!forceoverwrite
				  && cmd!='d' && !lstat(fullpath, &checkstat))
				{
					char msg[512]="";
printf("got lstat on restore\n");
					// Something exists at that path.
					snprintf(msg, sizeof(msg),
						"Path exists: %s (%s)",
							fullpath, dpth);
					if(restore_interrupt(cmd, dpth, msg))
					{
						ret=-1;
						quit++;
					}
					else
					{
						if(dpth) free(dpth);
						if(fullpath) free(fullpath);
						if(statbuf) { free(statbuf); statbuf=NULL; }
						if(buf) free(buf);
						buf=NULL;
						continue;
					}
				}
				break;	
			default:
				break;
		}
		if(dpth) free(dpth);

		if(!quit && !ret) switch(cmd)
		{
			case 'd': // directory data
                                if(restore_dir(cmd, fullpath, &statp))
				{
					ret=-1;
					quit++;
				}
				break;
			case 'f': // file data
				if(restore_file(cmd, fullpath, &statp))
				{
					logp("restore_file error\n");
					ret=-1;
					quit++;
				}
				break;
			case 'l': // symlink
			case 'L': // hardlink
				if(restore_link(cmd, fullpath, len,
					&statp, restoreprefix))
				{
					ret=-1;
					quit++;
				}
				break;
			case 's': // special file
				if(restore_special(cmd, fullpath, &statp))
				{
					ret=-1;
					quit++;
				}
				break;
			default:
				logp("unknown cmd: %c:%s\n", cmd, buf);
				quit++; ret=-1;
				break;
		}

		if(fullpath) free(fullpath);
		if(statbuf) { free(statbuf); statbuf=NULL; }
		if(buf) free(buf);
		buf=NULL;
	}
	if(statbuf) free(statbuf);

	if(!ret)
	{
		end_filecounter_restore(1);
		logp("restore finished\n");
	}
	else logp("ret: %d\n", ret);
	return ret;
}

/*
static void sighandler(int sig)
{
	logp("got signal: %d\n", sig);
	// Close the sockets properly so as to avoid annoying waits during
	// testing when I kill the server with a Ctrl-C and then get
	// 'unable to bind listening socket'.
	async_free();
	logp("exiting\n");
	exit(1);
}
*/

static void setup_signals(void)
{
	// Ignore SIGPIPE - we are careful with read and write return values.
#ifndef HAVE_WIN32
	//signal(SIGABRT, &sighandler);
	//signal(SIGTERM, &sighandler);
	//signal(SIGINT, &sighandler);
#endif
}

int client(struct config *conf, enum action act, const char *backup, const char *restoreprefix, const char *regex, int forceoverwrite)
{
	int ret=0;
	int rfd=-1;

	setup_signals();
//	settimers(0, 100);
	//logp("begin client\n");

	if((rfd=init_socket(conf->server, conf->port))<0) return 1;
	//logp("after init_socket\n");

	if(!(ret=async_init(rfd))
	 && !(ret=say_hello(conf)))
	{
		switch(act)
		{
			case ACTION_BACKUP:
				ret=do_backup(conf);
				break;
			case ACTION_RESTORE:
				ret=do_restore(conf, backup,
					restoreprefix, regex, forceoverwrite);
				break;
			case ACTION_LIST:
			case ACTION_LONG_LIST:
			default:
				ret=do_list(backup, regex, act);
				break;
		}
	}
	close_fd(&rfd);

        //logp("end client\n");
	return ret;
}
